/* IOT Copyright (C) 2010 CEG development team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Authors: CEG <ceg@ionlythink.com>, http://www.ionlythink.com
 */

/*********************************/
#include <algorithm>
#include <QApplication>
#include <QDir>
#include <QDomDocument>
/*********************************/
#include "MainController.h"
/*********************************/
#include "Box.h"
#include "Layer.h" // debug
#include "Window.h" // debug
#include "BoxController.h"
#include "IAction.h"
#include "Utils.h"
#include "WindowSystem.h"
#include "PopMenuAction.h"
#include "Logger.h"
/*********************************/

/************************************************* [ CTOR/DTOR ] *************************************************/

MainController::MainController(Systray & systray) :
        _view(*this, systray), _scenes(), _boxController(),
	_comGs(new WindowSystem), isInit(false)
{
	initializeXMLProfils();
}

MainController::~MainController()
{
    std::for_each(this->_scenes.begin(), this->_scenes.end(), Ceg::DeleteObject());
    delete this->_comGs;
}

/************************************************* [ GETTERS ] *************************************************/

BoxController const &	MainController::getBoxController() const
{
    return (this->_boxController);
}

ICommunicationGraphicalServer *	MainController::getComGs() const
{
    return (this->_comGs);
}


AbstractScene *	MainController::getCurrentScene() const
{
    return (this->_scenes.front());
}

const Profil&	MainController::getCurrentProfil()
{
	return (this->_currentProfil);
}

AbstractScene * MainController::getScene(const QString &id) const
{
    for (QList<AbstractScene *>::const_iterator it = this->_scenes.begin(), itEnd = this->_scenes.end(); it != itEnd; ++it)
    {
        if (**it == id)
        {
            return (*it);
        }
    }
    return (0);
}

int MainController::getIndexOfScene(AbstractScene * sceneSearch) const
{
    return (this->_scenes.indexOf(sceneSearch));
}

int MainController::getFirstOfType(BoxType type) const
{
    int i = 0;
    for (QList<AbstractScene *>::const_iterator it = this->_scenes.begin(), itEnd = this->_scenes.end(); it != itEnd; ++it, ++i)
    {
        if (**it == type)
        {
            return (i);
        }
    }
    return (-1);
}

AbstractScene * MainController::getSceneAt(int position) const
{
    if (position < this->_scenes.size())
    {
        return (this->_scenes.at(--position));
    }
    return (0);
}

AbstractScene * MainController::getFirstNavigationScene() const
{
    for (QList<AbstractScene *>::const_iterator it = this->_scenes.begin(), itEnd = this->_scenes.end(); it != itEnd; ++it)
    {
        if (**it == DEFAULT_BOX || **it == CUSTOM_BOX)
        {
            return (*it);
        }
    }
    return (0);
}

View &	MainController::getView()
{
    return (this->_view);
}

Ceg::Window & MainController::getWindow()
{
    return this->_ownWindow;
}

list<Profil>&	MainController::getProfils()
{
	return (this->_profils);
}

const Profil&	MainController::getProfil(const QString& name) const
{
	std::list<Profil>::const_iterator	it;

	for (it = this->_profils.begin(); it != this->_profils.end(); it++)
	{
		if ((*it)._name == name)
			return (*(it));
	}
}

/************************************************* [ SETTERS ] *************************************************/

void	MainController::setProfils(list<Profil>& profils)
{
	this->_profils = profils;
}

void	MainController::setProfil(const Profil& profil)
{
	this->_profils.push_back(profil);
}

void	MainController::setCurrentProfil(Profil profil)
{
	this->_currentProfil = profil;
}

/************************************************* [ OTHERS ] *************************************************/



void	MainController::initializeXMLProfils()
{

	QStringList	pathsToSearch;

#if defined(Q_OS_UNIX) || defined(Q_OS_MAC)
    pathsToSearch << "/usr/share";
    pathsToSearch << QCoreApplication::applicationDirPath() + "/share";
	static const QString osName("unix");
#elif defined(Q_OS_WIN)
	static const QString osName("win");
    pathsToSearch << QCoreApplication::applicationDirPath() + "/../share";
    pathsToSearch << QCoreApplication::applicationDirPath() + "/../";
#endif
	
	for (QDomNode boxNode = this->getBoxController()._appMenu.firstChild(); !boxNode.isNull(); boxNode = boxNode.nextSibling())
	{
		QString text;
		QString path;
		QDomElement const & boxElement = boxNode.toElement();
		for (QDomNode domNode = boxElement.firstChild(); !domNode.isNull(); domNode = domNode.nextSibling())
		{
			QDomElement const & childElement = domNode.toElement();
			if (childElement.isNull() == false)
			{
				QString const & tagName = childElement.tagName();
				if (childElement.firstChild().toElement().attribute("text") != "Back")
				{
					if (tagName == "action")
						path = childElement.firstChild().toElement().attribute(osName + "-path");
				}
					if (tagName == "style")
						text = childElement.attribute("text");
			}
		}
		if (path.isEmpty() == false)
			this->_appListing.insert(text, path);
	}

    for (QStringList::const_iterator it = pathsToSearch.constBegin(), itEnd = pathsToSearch.end(); it != itEnd; ++it)
    {
		QDir    directory(*it + "/" + "profils");
		if (directory.exists() == true)
		{
			this->_sharePath = *it;
			QFileInfoList const & files = directory.entryInfoList(QStringList("*.xml"));
			for (QFileInfoList::const_iterator it = files.begin(), itEnd = files.end();
			it != itEnd; ++it)
				this->loadXMLProfils(it->absoluteFilePath());
			break;
		}
    }
}

void				MainController::loadXMLProfils(QString const & fileName)
{
    QFile			file(fileName);
    QDomDocument	doc(fileName);
	
	if (file.open(QIODevice::ReadOnly) == true)
	{
		if (doc.setContent(&file) == true)
		{
			file.close();
			QDomElement const & rootElement = doc.documentElement();
			if (rootElement.tagName() == "profil")
			{
				Profil	profil(rootElement.attribute("name"));
				for (QDomNode shortcutNode = rootElement.firstChild(); !shortcutNode.isNull(); shortcutNode = shortcutNode.nextSibling())
				{
					QDomElement const & shortcutElement = shortcutNode.toElement();
					if (shortcutElement.isNull() == false && shortcutElement.tagName() == "shortcut")
						profil._shortcuts.insert(shortcutElement.attribute("key").toInt(), shortcutElement.attribute("action").toInt());
					else if (shortcutElement.isNull() == false)
						profil.setVrpnIp(shortcutElement.attribute("ip"));
				}
				this->_profils.push_back(profil);
			}
		}
	}
}


void    MainController::on_action_emitted(IAction & anAction)
{
    anAction.exec(*this);
}

AbstractScene * MainController::createScene(Ceg::Window const & window)
{
    Layer * oneLayer = new Layer(window, *this);
    QList<QGraphicsRectItem *> graphicItems;
    this->_boxController.getPattern(window, graphicItems);
    oneLayer->initialize(graphicItems);
    return (oneLayer);
}

void MainController::initialize()
{
	std::string	vrpnAdress = "openvibe-vrpn";
	vrpnAdress += this->_currentProfil._vrpnIp.toStdString();
    Logger::getInstance()->log(INFO_LOG, "Initialisation du MainController.");
    PopMenuAction popAction("Home");
    this->on_action_emitted(popAction);

    _running = true;
	_VRPNButton = new vrpn_Button_Remote(vrpnAdress.c_str());
    _VRPNButton->register_change_handler(/*&_running*/this, &(this->vrpn_button_callback));
    _timer = new QTimer(this);
    connect(_timer, SIGNAL(timeout()), this, SLOT(getVRPNData()));
    _timer->start();
}


void MainController::on_start_navigation()
{
    Logger::getInstance()->log(INFO_LOG,"Demarrage de la navigation.");
    if (isInit == false || this->_scenes.empty() == true) // si c est la premiere fois que l on lance la navigation ou qu'il n'y a plus de scene en cours d'utilisation
    {
	this->initialize();
	this->_view.initialize();
	isInit = true;
    }
    this->_view.show();
    this->_comGs->getFocusedWindow(this->_ownWindow);
}

void MainController::on_stop_navigation()
{
    Logger::getInstance()->log(INFO_LOG,"Arret de la navigation.");
    on_stop_eyetracking();
    this->_view.hide();
}

/************************************************************************/
/* Les Actions de l'eye-tracking (JEREM)                */
/************************************************************************/
void	MainController::Slt_DisplayHelp(bool status)
{
  _eyeTrackerGUI->Slt_DisplayHelp(status);
  _isInTrainingProcess = status;
}

void	MainController::Slt_BlinkDouble()
{
  if (!_isInTrainingProcess)
    _eyesActionsExec->Slt_BlinkDouble();
}

void	MainController::Slt_BlinkLeft()
{
  if (!_isInTrainingProcess)
    _eyesActionsExec->Slt_BlinkLeft();
}

void	MainController::Slt_BlinkRight()
{
  if (!_isInTrainingProcess)
    _eyesActionsExec->Slt_BlinkRight();
}

void	MainController::Slt_LookLeft()
{
  _eyeTrackerGUI->Update(LEFT);
  if (!_isInTrainingProcess)
    _eyesActionsExec->Slt_LookLeft();
}

void	MainController::Slt_LookBottom()
{
  _eyeTrackerGUI->Update(BOTTOM);
  if (!_isInTrainingProcess)
    _eyesActionsExec->Slt_LookBottom();
}

void	MainController::Slt_LookTop()
{
  _eyeTrackerGUI->Update(TOP);
  if (!_isInTrainingProcess)
    _eyesActionsExec->Slt_LookTop();
}

void	MainController::Slt_LookRight()
{
  _eyeTrackerGUI->Update(RIGHT);
  if (!_isInTrainingProcess)
    _eyesActionsExec->Slt_LookRight();
}

void	MainController::Slt_LookUnknown()
{
  _eyeTrackerGUI->Update(UNKNOWN);
  if (!_isInTrainingProcess)
    _eyesActionsExec->Slt_LookUnknown();
}



void MainController::on_start_eyetracking()
{
  _eyeTrackerThread= new QThreadEx;
  _eyeTracker= new EyeTrack(_eyesActions, _sharePath);
  _eyeTrackerGUI= new Gui(_sharePath, this);
  _eyesActionsExec= new eyes_actions_callback_func(this);
  _isInTrainingProcess= false;

  connect(&_eyesActions, SIGNAL(Sig_BlinkLeft()), this, SLOT(Slt_BlinkLeft()));
  connect(&_eyesActions, SIGNAL(Sig_BlinkRight()), this, SLOT(Slt_BlinkRight()));
  connect(&_eyesActions, SIGNAL(Sig_BlinkDouble()), this, SLOT(Slt_BlinkDouble()));
  connect(&_eyesActions, SIGNAL(Sig_LookLeft()), this, SLOT(Slt_LookLeft()));
  connect(&_eyesActions, SIGNAL(Sig_LookRight()), this, SLOT(Slt_LookRight()));
  connect(&_eyesActions, SIGNAL(Sig_LookBottom()), this, SLOT(Slt_LookBottom()));
  connect(&_eyesActions, SIGNAL(Sig_LookTop()), this, SLOT(Slt_LookTop()));
  connect(&_eyesActions, SIGNAL(Sig_LookUnknown()), this, SLOT(Slt_LookUnknown()));
  connect(&_eyesActions, SIGNAL(Sig_DisplayHelp(bool)), this, SLOT(Slt_DisplayHelp(bool)));
  connect(&_eyesActions, SIGNAL(Sig_No_Device()), _eyeTrackerGUI, SLOT(Slt_DisplayNoDevice()));
  connect(&_eyesActions, SIGNAL(Sig_EyesProblem(bool)), _eyeTrackerGUI, SLOT(Slt_DisplayEyesProblem(bool)));

  _eyeTracker->moveToThread(_eyeTrackerThread);
  connect(_eyeTrackerThread, SIGNAL(started()), _eyeTracker, SLOT(run_tracker()));
  connect(_eyeTracker, SIGNAL(finished()), _eyeTrackerThread, SLOT(quit()));
  connect(_eyeTracker, SIGNAL(finished()), _eyeTracker, SLOT(deleteLater()));
  connect(_eyeTrackerThread, SIGNAL(finished()), _eyeTrackerThread, SLOT(deleteLater()));
  connect(_eyeTrackerThread, SIGNAL(finished()), this, SLOT(cleanup_eyetracking()));
  _eyeTrackerThread->start();
}


void MainController::cleanup_eyetracking()
{
  delete _eyeTrackerGUI;
  delete _eyesActionsExec;
}

void MainController::on_stop_eyetracking()
{
  _eyeTracker->stop_tracker();
}

void MainController::pushFrontScene(AbstractScene *scene)
{
    this->_scenes.removeOne(scene); // Lorsqu'on effectue un zoom, cela evite d'avoir a deplacer la scene, on remove et au insert au depart
    this->_scenes.push_front(scene);
    this->_view.setScene(this->_scenes.front());
    this->_view.setSceneRect(this->_scenes.front()->getGeometry());
}

void MainController::popFrontScene()
{
    AbstractScene const * oldCurrentScene = this->_scenes.front();
    this->_scenes.pop_front();
    if (this->_scenes.empty() == false)
    {
        this->_view.setScene(this->_scenes.front());
        this->_view.setSceneRect(this->_scenes.front()->getGeometry());
    }
    delete oldCurrentScene;
}

void MainController::removeSceneAt(int index)
{
    AbstractScene const * deletedScene = this->_scenes.at(index);
    this->_scenes.removeAt(index);
    delete deletedScene;
    this->pushFrontScene(this->getScene("Home"));
}

// NIRLO
void MainController::getVRPNData()
{
  _VRPNButton->mainloop();
}

void VRPN_CALLBACK MainController::vrpn_button_callback(void* user_data, vrpn_BUTTONCB button)
{
  MainController* c = (MainController*)user_data;
  //std::cout << "Button ID : " << button.button << " / Button State : " << button.state << std::endl;
  int k = -1;
  int keyState = button.state == 1 ? 0 : 1;

  std::cout << button.button << std::endl;
  switch (button.button)
    {
    case 0: //RIGHT
#ifdef _WIN32
      k = VK_RIGHT;
#else
      k = XK_Right;
#endif
      break;
    case 1: //DOWN
#ifdef _WIN32
      k = VK_DOWN;
#else
      k = XK_Down;
#endif
      break;
    case 2: //RETURN
#ifdef _WIN32
      k = VK_RETURN;
#else
      k = XK_Return;
#endif
      break;
    }

  if (k == -1)
    {
      std::cerr << "Invalid key value!" << std::endl;
      return ;
    }
#ifdef _WIN32
  keyState == 0 ? keybd_event(k, 0, keyState, 0) : keyState = 0;
#else
  keyState == 0 ? c->_xAction.keyPress(k) : c->_xAction.keyRelease(k);
#endif
}
// !NIRLO
